;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "treepy" "0.1.3"
  "Generic tree traversal tools."
  '((emacs "25.1"))
  :url "https://github.com/volrath/treepy.el"
  :commit "28f0e2c2c75ea186e8beb570a4a70087926ff80b"
  :revdesc "28f0e2c2c75e"
  :keywords '("lisp" "maint" "tools")
  :authors '(("Daniel Barreto" . "daniel.barreto.n@gmail.com"))
  :maintainers '(("Daniel Barreto" . "daniel.barreto.n@gmail.com")))
